#pragma once

#include "pbo/pbo_problem.hpp"
#include "pbo/concatenated_trap.hpp"
#include "pbo/ising_ring.hpp"
#include "pbo/ising_torus.hpp"
#include "pbo/ising_triangular.hpp"
#include "pbo/labs.hpp"
#include "pbo/leading_ones.hpp"
#include "pbo/leading_ones_dummy1.hpp"
#include "pbo/leading_ones_dummy2.hpp"
#include "pbo/leading_ones_epistasis.hpp"
#include "pbo/leading_ones_neutrality.hpp"
#include "pbo/leading_ones_ruggedness1.hpp"
#include "pbo/leading_ones_ruggedness2.hpp"
#include "pbo/leading_ones_ruggedness3.hpp"
#include "pbo/linear.hpp"
#include "pbo/mis.hpp"
#include "pbo/nk_landscapes.hpp"
#include "pbo/n_queens.hpp"
#include "pbo/one_max.hpp"
#include "pbo/one_max_dummy1.hpp"
#include "pbo/one_max_dummy2.hpp"
#include "pbo/one_max_epistasis.hpp"
#include "pbo/one_max_neutrality.hpp"
#include "pbo/one_max_ruggedness1.hpp"
#include "pbo/one_max_ruggedness2.hpp"
#include "pbo/one_max_ruggedness3.hpp"



