#pragma once

#include "problem/transformation.hpp"
#include "problem/problem.hpp"
#include "problem/single.hpp"
#include "problem/wrap_function.hpp"
#include "problem/bbob.hpp"
#include "problem/pbo.hpp"
#include "problem/wmodel.hpp"
#include "problem/submodular.hpp"

#ifdef USING_MKLANDSCAPE
#include "problem/mklandscape/cliqueTreeC.hpp"
#endif

#include "problem/star_discrepancy/real.hpp"
#include "problem/star_discrepancy/integer.hpp"
