#pragma once

#include "bbob/bbob_problem.hpp" 
#include "bbob/sphere.hpp"
#include "bbob/ellipsoid.hpp"
#include "bbob/rastrigin.hpp"
#include "bbob/bueche_rastrigin.hpp"
#include "bbob/linear_slope.hpp"
#include "bbob/attractive_sector.hpp"
#include "bbob/step_ellipsoid.hpp"
#include "bbob/rosenbrock.hpp"
#include "bbob/rosenbrock_rotated.hpp"
#include "bbob/ellipsoid_rotated.hpp"
#include "bbob/discus.hpp"
#include "bbob/bent_cigar.hpp"
#include "bbob/sharp_ridge.hpp"
#include "bbob/different_powers.hpp"
#include "bbob/rastrigin_rotated.hpp"
#include "bbob/weierstrass.hpp"
#include "bbob/schaffers10.hpp"
#include "bbob/schaffers1000.hpp"
#include "bbob/griewank_rosenbrock.hpp"
#include "bbob/schwefel.hpp"
#include "bbob/gallagher101.hpp"
#include "bbob/gallagher21.hpp"
#include "bbob/katsuura.hpp"
#include "bbob/lunacek_bi_rastrigin.hpp"
#include "bbob/many_affine.hpp"